import copy
import dataclasses
import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional, TypeVar, cast

from camel_converter import to_snake

from validio_sdk.exception import ValidioError
from validio_sdk.graphql_client.exceptions import GraphQLClientHttpError
from validio_sdk.graphql_client.fragments import CredentialSecretChanged
from validio_sdk.resource._diff_util import must_find_source
from validio_sdk.resource._diffable import Diffable
from validio_sdk.resource._errors import (
    ManifestConfigurationError,
    max_resource_depth_exceeded,
    updated_resource_type_mismatch_exception,
)
from validio_sdk.resource._field_selector import SelectorWithFieldName
from validio_sdk.resource._resource import DiffContext, Resource
from validio_sdk.resource._util import SourceSchemaReinference, _sanitize_error
from validio_sdk.resource.channels import Channel
from validio_sdk.resource.credentials import Credential
from validio_sdk.resource.replacement import (
    CascadeReplacementReason,
    ImmutableFieldReplacementReason,
    ReplacementContext,
    ReplacementReason,
)
from validio_sdk.resource.sources import Source
from validio_sdk.validio_client import ValidioAPIClient

if TYPE_CHECKING:
    from validio_sdk.code.plan import ResourceNames

R = TypeVar("R", bound=Resource)

"""
When we descend into nested objects, we set a limit on how deep we go.
Otherwise, in a bad manifest configuration or due to a bug in our code, we
could enter a cycle.
"""
MAX_RESOURCE_DEPTH = 15


@dataclass
class ResourceUpdate:
    """
    Represents a resource that has an update. It contains the concrete class
    of the resource (e.g. TumblingWindow), followed by a representation of the
    manifest and server's version of the resource respectively.
    """

    manifest: Resource
    server: Resource
    # If there was an update to an immutable field on the resource,
    # consequently flagging the resource to be replaced, this field
    # contains the updated field.
    replacement_field: Optional[str] = None
    secrets_changed: bool = False


@dataclass
class ResourceUpdates:
    """Resources to update. Grouped by type."""

    credentials: dict[str, ResourceUpdate] = dataclasses.field(default_factory=dict)
    channels: dict[str, ResourceUpdate] = dataclasses.field(default_factory=dict)
    sources: dict[str, ResourceUpdate] = dataclasses.field(default_factory=dict)
    windows: dict[str, ResourceUpdate] = dataclasses.field(default_factory=dict)
    segmentations: dict[str, ResourceUpdate] = dataclasses.field(default_factory=dict)
    validators: dict[str, ResourceUpdate] = dataclasses.field(default_factory=dict)
    notification_rules: dict[str, ResourceUpdate] = dataclasses.field(
        default_factory=dict
    )


@dataclass
class GraphDiff:
    to_create: DiffContext
    to_delete: DiffContext
    to_update: ResourceUpdates

    replacement_ctx: ReplacementContext

    def retain(self, targets: "ResourceNames") -> None:
        """Retain resources.

        If a set of targets is specified, we only keep the resources that
        matches. If no targets are specified all will be targeted as usual.

        :param targets: `ResourceNames` with targets to preserve, ignoring the
            rest
        """
        if targets.size() == 0:
            return

        resource_types = DiffContext.fields()

        for resource_type in resource_types:
            resource_targets = getattr(targets, resource_type)
            to_create = getattr(self.to_create, resource_type)
            to_delete = getattr(self.to_delete, resource_type)
            to_update = getattr(self.to_update, resource_type)

            self._retain_targets(resource_targets, to_create)
            self._retain_targets(resource_targets, to_delete)
            self._retain_targets(resource_targets, to_update)

    @staticmethod
    def _retain_targets(targets: set[str], resources: dict[str, Any]) -> None:
        if len(resources) == 0:
            return

        to_remove = {r for r in resources if r not in targets}

        for resource_name in to_remove:
            del resources[resource_name]

    def num_creates(self) -> int:
        return self.size(self.to_create)

    def num_deletes(self) -> int:
        return self.size(self.to_delete)

    def num_updates(self) -> int:
        return self.size(self.to_update)

    def num_operations(self) -> int:
        return self.num_creates() + self.num_deletes() + self.num_updates()

    @staticmethod
    def size(obj: Any) -> int:
        count = 0
        for f in DiffContext.fields():
            count += len(getattr(obj, f))
        return count


async def diff_resource_graph(
    namespace: str,
    client: ValidioAPIClient,
    schema_reinference: SourceSchemaReinference,
    show_secrets: bool,
    manifest_ctx: DiffContext,
    server_ctx: DiffContext,
) -> GraphDiff:
    partial_diff = _diff_resource_graph(
        namespace=namespace,
        manifest_ctx=manifest_ctx,
        server_ctx=server_ctx,
        for_validators=False,
    )

    # Enrich resource_graph. We do this before expanding field selectors
    # because the latter depends on the schema for all sources being present.
    await enrich_resource_graph(
        manifest_ctx=manifest_ctx,
        server_ctx=server_ctx,
        graph=partial_diff,
        client=client,
        schema_reinference=schema_reinference,
        show_secrets=show_secrets,
    )

    # Expand field selector for validators.
    expand_validator_field_selectors(manifest_ctx)

    # Now after expanding field selectors, we can diff validators.
    validators_diff = _diff_resource_graph(
        namespace=namespace,
        manifest_ctx=manifest_ctx,
        server_ctx=server_ctx,
        for_validators=True,
    )

    final_diff = partial_diff
    final_diff.to_create.validators = validators_diff.to_create.validators
    final_diff.to_update.validators = validators_diff.to_update.validators
    final_diff.to_delete.validators = validators_diff.to_delete.validators
    for k, v in validators_diff.replacement_ctx.validators.items():
        final_diff.replacement_ctx.validators[k] = v

    _compute_replacements(
        manifest_ctx=manifest_ctx,
        server_ctx=server_ctx,
        graph_diff=final_diff,
    )

    return final_diff


def _compute_replacements(
    manifest_ctx: DiffContext,
    server_ctx: DiffContext,
    graph_diff: GraphDiff,
) -> None:
    resource_topo_order = DiffContext.fields_topological_order(
        graph_mode=True,
    )

    for parent_resource_type, _ in resource_topo_order:
        to_update: dict[str, ResourceUpdate] = getattr(
            graph_diff.to_update, parent_resource_type
        )
        replacement_ctx: dict[str, ReplacementReason] = getattr(
            graph_diff.replacement_ctx, parent_resource_type
        )

        to_replace = {
            name: (resource_update, resource_update.replacement_field)
            for name, resource_update in to_update.items()
            if resource_update.replacement_field
        }

        for name, (resource_update, replacement_field) in to_replace.items():
            replacement_ctx[name] = ImmutableFieldReplacementReason(
                field_name=replacement_field, resource_update=resource_update
            )
            _visit_resource_to_replace(
                manifest_ctx=manifest_ctx,
                server_ctx=server_ctx,
                graph_diff=graph_diff,
                resource_type=parent_resource_type,
                resource_name=name,
                curr_depth=1,
            )


def _visit_resource_to_replace(
    manifest_ctx: DiffContext,
    server_ctx: DiffContext,
    graph_diff: GraphDiff,
    resource_type: str,
    resource_name: str,
    curr_depth: int,
) -> None:
    if curr_depth > MAX_RESOURCE_DEPTH:
        raise ManifestConfigurationError(
            "BUG: max recursion depth exceeded while computing replacements"
        )

    to_update: dict[str, ResourceUpdate] = getattr(graph_diff.to_update, resource_type)
    to_create: dict[str, Resource] = getattr(graph_diff.to_create, resource_type)
    to_delete: dict[str, Resource] = getattr(graph_diff.to_delete, resource_type)

    # Remove the resource from the update list.
    if resource_name in to_update:
        del to_update[resource_name]

    # Flag the resource to be created.
    manifest_resources: dict[str, Resource] = getattr(manifest_ctx, resource_type)
    manifest_obj = manifest_resources[resource_name]
    to_create[manifest_obj.name] = manifest_obj

    # If the resource exists on the server, flag it resource to be deleted.
    server_resources: dict[str, Resource] = getattr(server_ctx, resource_type)
    if resource_name in server_resources:
        server_obj = server_resources[resource_name]
        to_delete[server_obj.name] = server_obj

    # Check for cascading child resources to replace
    # We only need to look at what's in the manifest (resources
    # not in the manifest are already being flagged for deletion).
    resource_topo_order = DiffContext.fields_topological_order(graph_mode=True)
    # Get all children resources for the resource and flag them to be moved as well.
    children_types = next(
        (
            children_types
            for (parent_type, children_types) in resource_topo_order
            if parent_type == resource_type
        ),
        None,
    )
    if not children_types:
        return

    parent_field_name = f"{manifest_obj.resource_class_name().lower()}_name"
    for child_resource_type in children_types:
        child_resources: dict[str, Resource] = getattr(
            manifest_ctx, child_resource_type
        )
        replacement_ctx: dict[str, ReplacementReason] = getattr(
            graph_diff.replacement_ctx, child_resource_type
        )

        for child_resource_name, child_resource in child_resources.items():
            if resource_name != getattr(child_resource, parent_field_name):
                continue
            replacement_ctx[child_resource_name] = CascadeReplacementReason(
                parent_resource_cls=manifest_obj.__class__,
                parent_resource_name=resource_name,
            )

            _visit_resource_to_replace(
                manifest_ctx=manifest_ctx,
                server_ctx=server_ctx,
                graph_diff=graph_diff,
                resource_type=child_resource_type,
                resource_name=child_resource_name,
                curr_depth=curr_depth + 1,
            )


def expand_validator_field_selectors(manifest_ctx: DiffContext) -> None:
    for name in list(manifest_ctx.validators.keys()):
        template = manifest_ctx.validators[name]

        source_field_selector: SelectorWithFieldName | None = getattr(
            template, "_field_selector", None
        )
        reference_field_selector: dict[str, str] | None = getattr(
            template, "_reference_field_selector", None
        )
        filter_field_selector: SelectorWithFieldName | None = (
            getattr(template.filter, "_field_selector", None)
            if template.filter
            else None
        )

        if source_field_selector:
            field_selector = source_field_selector
        elif filter_field_selector:
            field_selector = filter_field_selector
        else:
            continue

        source = must_find_source(manifest_ctx, template.source_name)
        if source.jtd_schema is None:
            # If we have no schema then it means we're doing a diff where the
            # source's credential has not yet been created - which means we can't
            # do schema inference. As a result, we can't yet expand the selector.
            continue

        if source_field_selector:
            delattr(template, "_field_selector")
        if reference_field_selector:
            delattr(template, "_reference_field_selector")
        if filter_field_selector:
            delattr(template.filter, "_field_selector")

        # Remove the placeholder validator.
        del manifest_ctx.validators[name]

        fields = field_selector.field_selector._get_matching_fields(
            cast(Any, source.jtd_schema)
        )

        # We don't want to deep copy the entire resource graph
        resource_graph = template._resource_graph
        template._resource_graph = None  # type: ignore
        for field in fields:
            validator = copy.deepcopy(template)
            validator._resource_graph = resource_graph

            # We must obey to our own naming convention so we drop everything
            # that's not valid.
            resource_name = validator.name % {"field": field}
            resource_name = resource_name.replace(" ", "_").lower()
            resource_name = re.sub(r"[^a-z0-9_.-]", "_", resource_name)
            validator.name = resource_name

            if source_field_selector:
                setattr(validator, field_selector.field_name, field)

            if reference_field_selector:
                setattr(validator, reference_field_selector["field_name"], field)

            if filter_field_selector:
                setattr(
                    validator.filter,
                    field_selector.field_name,
                    field,
                )

            manifest_ctx.validators[resource_name] = validator


def _diff_resource_graph(
    namespace: str,
    manifest_ctx: DiffContext,
    server_ctx: DiffContext,
    for_validators: bool | None = None,
) -> GraphDiff:
    fns = [
        (compute_creates, DiffContext),
        (compute_deletes, DiffContext),
        (compute_updates, ResourceUpdates),
    ]
    diffs = []
    fields = DiffContext.fields()
    for diff_fn, cls in fns:
        diff_by_resource = {}
        for field in fields:
            include_diff = (
                True
                if for_validators is None
                else (field == "validators") == for_validators
            )
            if not include_diff:
                continue
            diff_by_resource[field] = diff_fn(
                namespace, getattr(manifest_ctx, field), getattr(server_ctx, field)
            )
        diffs.append(cls(**diff_by_resource))

    return GraphDiff(
        to_create=diffs[0],
        to_delete=diffs[1],
        to_update=diffs[2],
        replacement_ctx=ReplacementContext(),
    )


def compute_creates(
    namespace: str, manifest_resources: dict[str, R], server_resources: dict[str, R]
) -> dict[str, R]:
    creates = {}
    for name, resource in manifest_resources.items():
        if name in server_resources:
            _check_namespace(namespace, server_resources[name])
            continue
        creates[name] = resource
    return creates


def compute_deletes(
    namespace: str, manifest_resources: dict[str, R], server_resources: dict[str, R]
) -> dict[str, R]:
    deletes = {}
    for name, resource in server_resources.items():
        if namespace != resource._must_namespace():
            # Only work on resources in the configured namespace.
            continue

        if name not in manifest_resources:
            deletes[name] = resource

    return deletes


def compute_updates(
    namespace: str, manifest_resources: dict[str, R], server_resources: dict[str, R]
) -> dict[str, ResourceUpdate]:
    diffs = {}

    for name, manifest in manifest_resources.items():
        if name not in server_resources:
            continue

        server = server_resources[name]
        _check_namespace(namespace, server)

        d = diff(manifest, server, 0, manifest, server)
        if d:
            diffs[name] = d

    return diffs


# ruff: noqa: PLR0911,PLR0912
def diff(
    manifest_object: Diffable,
    server_object: Diffable,
    curr_depth: int,
    manifest_resource: Resource,
    server_resource: Resource,
) -> ResourceUpdate | None:
    """
    Compares the current (manifest) resource against a provided server version
    of that resource. None is returned if there are no changes between the two
    resources. If there is at least one change, then a diff is returned which
    contains a representation of the full manifest and server side resource.
    """
    if curr_depth > MAX_RESOURCE_DEPTH:
        raise max_resource_depth_exceeded(manifest_resource.name)

    # If the server resource has a different type, this is an invalid update.
    # e.g. we can't switch a window from tumbling to fixed batch. The window
    # needs to be re-created if that's the desired outcome. This only applies
    # for resources, nested objects like filters can still change type.
    if not isinstance(server_object, manifest_object.__class__):
        if isinstance(server_object, Resource):
            raise updated_resource_type_mismatch_exception(
                manifest_resource.name, manifest_object, server_object
            )
        return ResourceUpdate(manifest_resource, server_resource)

    # If we've been requested explicitly flagged to not diff the resource,
    # then no changes to report.
    if isinstance(manifest_object, Resource) and manifest_object.ignore_changes:
        return None

    # Check for updates to immutable fields. If we detect such changes,
    # we flag the resource to be recreated.
    for field in manifest_object._immutable_fields():
        manifest = getattr(manifest_object, field)
        server = getattr(server_object, field)
        if manifest != server:
            return ResourceUpdate(
                manifest_resource, server_resource, replacement_field=field
            )

    # If any valid field changes, then mark the resource as having a diff.
    for field in manifest_object._mutable_fields():
        manifest = getattr(manifest_object, field)
        server = getattr(server_object, field)

        if manifest != server:
            # This is a temporary fix to support marking secret fields as
            # changed on a `Channel`. If the field is both in `_mutable_fields`
            # and also in `_secret_fields` we will flag `secrets_changed` if
            # there has been an update. This ensures that the value is redacted
            # by default when displaying the config.
            secrets_changed = False
            if (
                isinstance(manifest_object, Channel)
                and field in manifest_object._secret_fields()
            ):
                secrets_changed = True

            # Whenever we find there's a change to make, grab all fields on both
            # resource versions so that we can present them as a diff.
            # Regardless of at what depth we find a change, the diff we present
            # will always start from the root resource itself. We pass those objects
            # around for this purpose.
            return ResourceUpdate(
                manifest_resource,
                server_resource,
                secrets_changed=secrets_changed,
            )

    # Some fields require the resource to be re-created if the type changes
    # e.g. going from a dynamic to a fixed threshold is not allowed.
    # If we detect that then we flag the resource to be re-created.
    for field in manifest_object._replace_on_type_change_fields():
        manifest = getattr(manifest_object, field)
        server = getattr(server_object, field)
        if server.__class__ != manifest.__class__:
            return ResourceUpdate(
                manifest_resource, server_resource, replacement_field=field
            )

    # No changes yet. Next, descend into the nested fields. If we find any
    # changes, then mark _this_ resource as changed.
    server_nested_objects = server_object._nested_objects()
    for field, manifest in manifest_object._nested_objects().items():
        server = server_nested_objects[field]

        # No possible change if both are unset.
        if manifest is None and server is None:
            continue

        # If exactly one of them is None, then we definitely have a change.
        if manifest is None or server is None:
            return ResourceUpdate(manifest_resource, server_resource)

        if isinstance(server, list) != isinstance(manifest, list):
            return ResourceUpdate(manifest_resource, server_resource)

        if isinstance(server, list) and isinstance(manifest, list):
            if len(server) != len(manifest):
                return ResourceUpdate(manifest_resource, server_resource)

            for i in range(len(server)):
                # Descend into both objects to diff.
                collected_diff = diff(
                    manifest[i],
                    server[i],
                    curr_depth + 1,
                    manifest_resource,
                    server_resource,
                )

                if collected_diff:
                    # The returned diff is that of the full resource (not
                    # just the nested object that had a change). So we can
                    # exit with it.
                    return collected_diff

            continue

        # Descend into both objects to diff.
        collected_diff = diff(
            manifest, server, curr_depth + 1, manifest_resource, server_resource
        )
        if collected_diff:
            # The returned diff is that of the full resource (not
            # just the nested object that had a change). So we can
            # exit with it.
            return collected_diff

    # No update to make
    return None


async def enrich_resource_graph(
    manifest_ctx: DiffContext,
    server_ctx: DiffContext,
    graph: GraphDiff,
    client: ValidioAPIClient,
    schema_reinference: SourceSchemaReinference,
    show_secrets: bool,
) -> None:
    """
    Now that we have a graph and done a diff, some values are unknown and need to
    be resolved before we can get a full picture of changes and start making api
    requests to create/update things. This function fetches missing info from the
    server and updates resources in the graph as needed.
    """
    # For the resources we update, resolve the ids in the manifest objects. We have
    # the info in the equivalent server object.
    for f in DiffContext.fields():
        manifest_resources = getattr(manifest_ctx, f)
        server_resources = getattr(server_ctx, f)
        for name, server_resource in server_resources.items():
            if name in manifest_resources:
                manifest_resource = manifest_resources[name]
                manifest_resource._id.value = server_resource._must_id()

    # Since source jtd schemas can be managed automatically, we do not include
    # them in the actual diff process since they are not necessarily known yet.
    # Instead, here we explicitly check for any updates.
    await check_for_source_schema_changes(
        manifest_ctx, server_ctx, graph, client, schema_reinference
    )

    # Check if any credential secrets has changed and add them to the diff.
    # They're excluded by default.
    await check_for_credential_secret_changes(
        manifest_ctx, server_ctx, graph, client, show_secrets
    )


async def check_for_source_schema_changes(
    manifest_ctx: DiffContext,
    server_ctx: DiffContext,
    graph: GraphDiff,
    client: ValidioAPIClient,
    schema_reinference: SourceSchemaReinference,
) -> None:
    # For sources that are to be created and don't have a schema, we
    # need to infer a schema for them to use in the create request.
    for name, source in graph.to_create.sources.items():
        if source.jtd_schema is None:
            await infer_schema_for_source(manifest_ctx, source, client)

    # For sources that are to be updated, if there is no schema specified
    # in the manifest, then assign it the schema of the server version.
    # Or if we were asked to re-infer the schema, then do that instead.
    for name, r in graph.to_update.sources.items():
        assert isinstance(r.manifest, Source)
        assert isinstance(r.server, Source)

        manifest_source = r.manifest
        reinfer = schema_reinference.should_reinfer_schema_for_source(name)
        if manifest_source.jtd_schema is None:
            if reinfer:
                await infer_schema_for_source(manifest_ctx, manifest_source, client)
            else:
                manifest_source.jtd_schema = r.server.jtd_schema

    # For unchanged sources, check if there is a schema diff now that all schemas
    # have been resolved. If we find a diff, flag the source as updated.
    unchanged_sources = {
        name: s
        for name, s in manifest_ctx.sources.items()
        if s.name not in graph.to_create.sources
        and s.name not in graph.to_update.sources
        and s.name not in graph.to_delete.sources
    }

    for name, server_source in server_ctx.sources.items():
        if name not in unchanged_sources:
            continue

        manifest_source = unchanged_sources[name]

        reinfer = schema_reinference.should_reinfer_schema_for_source(name)
        if manifest_source.jtd_schema is None and reinfer:
            await infer_schema_for_source(manifest_ctx, manifest_source, client)

        if (
            manifest_source.jtd_schema is not None
            and manifest_source.jtd_schema != server_source.jtd_schema
        ):
            # The 'manifest' and 'server' version of the resource can be the same here
            # since there is no difference between the two (only change is the secret).
            graph.to_update.sources[name] = ResourceUpdate(
                manifest_source, server_source
            )

        # Ensure the source has a schema in case we need it in a field selector.
        if manifest_source.jtd_schema is None:
            manifest_source.jtd_schema = server_source.jtd_schema


async def infer_schema_for_source(
    manifest_ctx: DiffContext,
    source: Source,
    client: ValidioAPIClient,
) -> None:
    credential = manifest_ctx.credentials[source.credential_name]
    # If we don't yet have an ID (credential has not yet been created),
    # we can't do schema inference. So schema will be unknown in the diff.
    if credential._id.value is None:
        return
    await source._api_infer_schema(credential, client)


async def check_for_credential_secret_changes(
    manifest_ctx: DiffContext,
    server_ctx: DiffContext,
    graph: GraphDiff,
    client: ValidioAPIClient,
    show_secrets: bool,
) -> None:
    """
    To be able to always compute if we have any diff for our credentials we want
    to call the API to figure this out otherwise we wouldn't show a change in
    the secret if other fields were changed.

    TODO(VR-3095): The API should return a list of changed fields so we can be
    more granular in our diff. Currently we will mark *all* secret fields as
    changed.
    """
    # If it's a delete it won't exist in `manifest_ctx` so it will be skipped.
    for name, manifest_resource in manifest_ctx.credentials.items():
        # If it's an add it won't exist in `server_ctx` so it will be skipped.
        server_resource = server_ctx.credentials.get(name)
        if not server_resource:
            continue

        if not _has_secret_fields(manifest_resource, curr_depth=1):
            continue

        has_changed = await _credential_secret_change(
            client, manifest_resource, show_secrets
        )

        if not has_changed:
            continue

        # Flag the credential for secrets changed if any so we can show it in
        # the diff.
        graph.to_update.credentials[name] = ResourceUpdate(
            manifest_resource, server_resource, secrets_changed=has_changed
        )


# Checks if the Diffable object or any of its sub objects
# defines any secret fields.
def _has_secret_fields(obj: Diffable, curr_depth: int) -> bool:
    if curr_depth > MAX_RESOURCE_DEPTH:
        raise ValidioError(
            "BUG: max recursion depth exceeded while looking for secrets"
        )

    if hasattr(obj, "_secret_fields") and obj._secret_fields():
        return True

    for nested in obj._nested_objects().values():
        if isinstance(nested, list):
            for n in nested:
                if _has_secret_fields(n, curr_depth=curr_depth + 1):
                    return True
        elif nested and _has_secret_fields(nested, curr_depth + 1):
            return True

    return False


async def _credential_secret_change(
    client: ValidioAPIClient,
    credential: Credential,
    show_secrets: bool = False,
) -> bool:
    if credential.ignore_changes:
        return False

    method_name = f"{to_snake(credential.__class__.__name__)}_secret_changed"

    try:
        response = cast(
            CredentialSecretChanged,
            await client.__getattribute__(method_name)(
                credential._api_credential_secret_change_input(),
            ),
        )
    except GraphQLClientHttpError as e:
        raise _sanitize_error(e, show_secrets)

    if len(response.errors) > 0:
        raise ValidioError(
            f"operation '{method_name}' failed for Credential {credential.name}: "
            f"{response.errors}"
        )

    return response.has_changed or False


def _check_namespace(namespace: str, server_resource: Resource) -> None:
    server_namespace = server_resource._must_namespace()
    if namespace != server_namespace:
        raise ManifestConfigurationError(
            f"resource {server_resource.__class__.__name__} '{server_resource.name}' "
            "does not belong to the current namespace; "
            f"resource namespace = {server_namespace}; current namespace = {namespace}"
        )
